/**
 * ===========================================================================
 *                     DEPTH FIRST VISIT OF A TREE 
 *
 * Grammar G for the input tree:  
 *                           char ::= '0' | '1' | ... | '9'
 *                           tree ::= char | - char tree | (tree char tree)
 * Internal tree:            tree = Num  |  Num x tree  |  tree x Num x tree
 *                                 (Leaf)     (Unary)          (Binary)
 * Grammar for the output tree: tree ::= n | (n.tree) | (tree.n.tree)
 * 
 * A tree is input as a string generated by the grammar G and then 
 * it is visited in the depth first manner.
 * We assume that the word 'node' is equivalent to the word 'tree'.
 * 
 * ------------
 * There are two version of the program: 
 * (i) an iterative version and (ii) a recursive version.
 * The iterative version is uncommented. In order to use the recursive version
 * one should comment the iterative version and uncomment the recursive one.
 * ===========================================================================
 */
package parser;

import parserUtil.ConsoleReader;
import parserUtil.Node;
import java.io.StringReader;
import java.io.IOException;

public class DepthFirstVisit {
    /**
     * This class transforms the input string into a tree of type Node.
     * The string encodes the tree as indicated by the grammar G.
     * @param string: the string which encodes the given tree
     * @return the root of the tree generated by the parsing
     * @throws IOException
     */
   public static Node parse(String string)
      throws IOException, ParseException {
      StringReader sr = new StringReader(string); // sr is a sequence of 
      return parse(sr);	                         // characters, that is, 
   }                                              // sr belongs to char*

	                         // this method overloads the above parse method
   public static Node parse(StringReader string)
      throws IOException, ParseException {
      char c = (char)string.read();
      if ( ('0' <= c ) && ( c <= '9') ) {
         return new LeafNode(new Integer(c - 48));
      } else
      if ( c == '-') {
         char c2        = (char)string.read();
         Integer value  = new Integer(c2 - 48);
         Node sonTree   = parse(string);
         UnaryNode node = new UnaryNode(value, sonTree);
         return node;
      } else
      if ( c == '(' ) {
         Node leftTree   = parse(string);
         char c2         = (char)string.read();
         Integer value   = new Integer(c2 - 48);
         Node rightTree  = parse(string);
         BinaryNode node = new BinaryNode(value, leftTree, rightTree);
         c2              = (char)string.read();
         return node;
      } else {
      throw new ParseException();
      }
   }
  /**
   * Depth first visit of the given tree 
   * @param tree : tree to be visited in a depth first manner
   * @return the sequence of nodes which is the depth first visit of 
   *         the given tree
   */
   public static String depthFirstVisit(Node tree) {
      TreeStack treestack = new TreeStack();
      StringBuffer buffer = new StringBuffer();  

// ---------------------- ITERATIVE VERSION ----------------------------------
                              // iteration which uses a stack of trees      //
      treestack.push(tree);                                                 //
      while(!treestack.isEmpty()) {                                         //
         Node topNode = treestack.pop();                                    //
         if ( topNode instanceof LeafNode ) {                               //
            buffer.append(((LeafNode)topNode).value + " ");                 //
        } else                                                              //
         if ( topNode instanceof UnaryNode ) {                              //
            LeafNode ln = new LeafNode(((UnaryNode)topNode).value);         //
            treestack.push(ln);                                             //
            treestack.push(((UnaryNode)topNode).getChild());                //
         } else                                                             //
         if ( topNode instanceof BinaryNode ) {                             //
            LeafNode ln = new LeafNode(topNode.value);                      //
            treestack.push(ln);                                             //
            treestack.push(((BinaryNode)topNode).getRightChild());          //
            treestack.push(((BinaryNode)topNode).getLeftChild());           //
         }                                                                  //
      }                                                                     //
// ---------------------- END OF ITERATIVE VERSION ---------------------------
/* 
// ---------------------- RECURSIVE VERSION ----------------------------------
                              // recursion which does not a stack of trees  //
 if ( tree instanceof LeafNode ) {                                          //
           buffer.append(((LeafNode)tree).value + " ");                     //
      } else                                                                //
      if ( tree instanceof UnaryNode ) {                                    //
           buffer.append(depthFirstVisit(((UnaryNode)tree).getChild())      //
                     + tree.value + " ");                                   //
      } else                                                                //
      if ( tree instanceof BinaryNode ) {                                   //
           buffer.append(depthFirstVisit(((BinaryNode)tree).getLeftChild()) //
                     + depthFirstVisit(((BinaryNode)tree).getRightChild())  //
                     + tree.value + " ");                                   //
      };                                                                    //
// ---------------------- END OF RECURSIVE VERSION ---------------------------
*/
      return buffer.toString();
   }
    
   /**
    * @param args : the arguments of the command line
    */
    public static void main(String[] args) {
       ConsoleReader console = new ConsoleReader(System.in);
       System.out.println("Input a tree according to the grammar given "
                         + "in the program, please."); 	
       String inputLine = console.readLine();
       Node tree = null;
       try {
           tree = parse(inputLine);
       } catch ( IOException e ) {
            e.printStackTrace();
       } catch ( ParseException pex ) {
            System.out.println(pex.getMessage());
       }
       if (tree != null) {
            System.out.println("\nTree in input:\n" + tree);
            System.out.println("\nDepth first visit of the tree:");
            System.out.println(depthFirstVisit(tree));
       }
   }
}
/** 
 * input:  output:
 * ---------------------------------------------------------------------------
 * javac -classpath . parser/DepthFirstVisit.java
 * java  -classpath . parser/DepthFirstVisit
 *
 *         Input a tree according to the grammar given in the program, please.
 * (-231(-564-78))
 *
 *         Tree in input:
 *         ((2.3).1.((5.6).4.(7.8)))
 *
 *         Depth first visit of the tree:
 *         3 2 6 5 8 7 4 1
 * ---------------------------------------------------------------------------
 * The string   (-231(-564-78))   encodes the tree:     1
 *                                                     / \
 *                                                     2  4
 *                                                     |  /\
 *                                                     3 5  7
 *                                                       |  |
 *                                                       6  8
 * whose depth first visit is: 3 2 6 5 8 7 4 1
 * ---------------------------------------------------------------------------
 */
