/**
 * ===========================================================================
 *                 SUMMING UP THE ELEMENTS OF A MATRIX
 *
 * Thanks to Prof. Marco Danelutto. 
 * With changes by Alberto Pettorossi.
 *
 * We are given an n x n matrix and we want to "reduce" all the elements of 
 * that matrix via an associative and commutative operator. For instance, if 
 * the associative and commutative operator is the addition, we want to  
 * compute the sum of all the elements of the given matrix. 
 * This summation is done row by row by n independent threads, each of them 
 * working at each row. A final thread will sum up the values of the partial 
 * sum as soon as they are computed.
 * 
 * We assume this class defining the associative and commutative operator:
 * 
 *  public class CommAssOperator {
 *     public static int sum(int a, int b) {
 *           return (a + b);
 *     }
 *  }
 *
 * The values of the elements of the matrix are generated by the
 * following code:
 *
 *  Random generator = new Random();
 *     for(int i=0;i < n;i++) {
 *        for(int j=0; j < n; j++) {
 *           matrix[i][j] = generator.nextInt(5);
 *        }
 *     }
 *
 * MatrixSumMain generates the matrix and the various threads.
 *
 * FinalSum sums up the partial sums of the rows. These sums
 * are placed in a queue as soon as they have been computed.
 * An exception stops this final summing thread when no more
 * partial sums are placed in the queue.
 *
 * The RowSumThread class is the class of the thread which sums
 * the elements of each row. Each threads of this class places the
 * partial sum in a queue. The queue allows n insertions and n
 * extractions only. The termination of the activities is controlled
 * by the exception mechanism.
 * ===========================================================================
 */

import java.util.*;

class CommAssOperator {
    public static int sum(int a, int b) {
       return (a + b);
    }
}

public class MatrixSumMain {

  public static void main(String [] args) {
      int n = Integer.parseInt(args[0]);
      int [][] matrix = new int [n][n];
      Random generator = new Random();
      int s = 0;
      for(int i=0;i<n;i++) {
        for(int j=0;j<n;j++) {
            matrix[i][j] = generator.nextInt(5);
            s = CommAssOperator.sum(matrix[i][j],s);
            System.out.print(matrix[i][j]+" "); 
        }
        System.out.println();
      }
      System.out.println("The sum of all the elements of the matrix is: "+ s);
      SumQueue queue = new SumQueue(n);
      
      FinalSum finSum = new FinalSum(queue);
      finSum.start();
      
      Thread [] adders = new Thread[n];
      for(int i=0;i<n;i++) {
        adders[i] = new RowSumThread(matrix,i,queue);
        adders[i].start();
      }   
      return;
  }    
}
/**
 * input:            output:
 * --------------------------------------------------------------------------- 
 * javac MatrixSumMain.java
 * java  MatrixSumMain 5  
 *
 *                   0 3 0 3 4
 *                   2 0 4 2 1
 *                   4 1 0 1 2
 *                   4 1 3 4 4
 *                   3 2 2 1 0
 *                   The sum of all the elements of the matrix is: 51
 *                   FinalSumThread Thread-0 started ...
 *                   RowSumThread Thread-1 started ...
 *                   RowSumThread Thread-1 ended. Sum of row 1 is: 10
 *                   RowSumThread Thread-2 started ...
 *                   RowSumThread Thread-2 ended. Sum of row 2 is: 9
 *                   RowSumThread Thread-3 started ...
 *                   RowSumThread Thread-3 ended. Sum of row 3 is: 8
 *                   RowSumThread Thread-4 started ...
 *                   RowSumThread Thread-4 ended. Sum of row 4 is: 16
 *                   RowSumThread Thread-5 started ...
 *                   RowSumThread Thread-5 ended. Sum of row 5 is: 8
 *                   FinalSumThread Thread-0 ended.
 *                   The computed sum of all the elements of the matrix is: 51
 * ---------------------------------------------------------------------------
 */
