/**
 * ==========================================================================
 *                  Main Program for Testing Mutual Exclusion
 *                        using Peterson's Algorithm
 *                     Version with one public class only.
 *
 * Thanks to Prof. Vijay K. Garg, Austin, Texas.
 * The N processes are P_0, P_1, ..., P_{N-1}. N is assumed to be >= 2.
 *
 * In order to test Peterson's algorithm with N processes P_0, P_1, ..., 
 * P_{N-1}, we type:
 * 
 *    javac PetersonNProcess.java
 *    java  PetersonNProcess N
 *
 * For i = 0, ..., N-1, in the output:
 *   i  means that process P_i is in its non-critical section
 *  [i] means that process P_i is in its     critical section
 * ------------------------------------------------
 * The indexes i and j which occur in the methods 
 *    takeLock(int i)    and 
 *    releaseLock(int i) 
 * belong to the set {0, 1, ..., N-1}.
 * ==========================================================================
 */

import java.util.Random;

public class PetersonNProcesses extends Thread {

// ---------------------------------------------------------------------------
 private static class Peterson {     // version for N processes

    int n;                           // the variable n holds the value of N
    int [] q ;      
    int [] s ; 

    private Peterson (int n) {       // constructor for N processes
    this.n = n;
    this.q = new int [n];
    this.s = new int [n-1];
                                     // initially: q[0] = ... = q[N-1] = 0
    for (int i=0; i<n; i++) {   
         this.q[i] = 0;                                 
      };                             // initially: s[0] = ... = s[N-2] = 1
    for (int i=0; i<n-1; i++) {                       
         this.s[i] = 1;    
      }; 
    }

    public void takeLock(int i){     // entry protocol

      for (int j=0; j<n-1; j++) {
         q[i] = j+1;
         s[j] = i+1;   

         boolean b;
         do { 
              b = true;
              for (int k=0; k < n ; k++) {
                 if (k != i) {b = b && (q[k] < (j+1));};
                 }; 
            }
         while (!b && (s[j] == (i+1)));
      }
    }                                // end of entry protocol

    public void releaseLock(int i) { // exit protocol  
      q[i] = 0;
    }                                // end of exit protocol
 };
// ---------------------------------------------------------------------------

   int id;
   Peterson peterson = null;

   public PetersonNProcesses (int id, Peterson peterson) {
      this.id = id;                  // id of process P_i is i
      this.peterson = peterson;      // for i= 0, ..., N-1 
   } 

   private static void Sleep(int time){
      try {
           Thread.sleep(time);
      }
      catch (InterruptedException e) {}
   }

   void nonCriticalSection() {       // the duration of the non-Critical
      System.out.print(" "+ id);     // Section is longer for processes
      Sleep(300 * id + 10);          // with higher id
   }                                   

   void CriticalSection(){           // the duration of the Critical Section
      System.out.print(" ["+ id);    // is the same for all processes
      Sleep(80);
      System.out.print("]"); 

   }

   public void run() {
      while (true) {
         nonCriticalSection();
         peterson.takeLock(id);  
         CriticalSection();
         peterson.releaseLock(id);      
      }
   }

   public static void main(String [] args) {

                               // N is the number of processes
      final int N = Integer.parseInt(args[0]); 

                               // t is an array of N threads
      PetersonNProcesses [] t = new PetersonNProcesses[N]; 
    
                               // only one Peterson's object is constructed
      Peterson peterson = new Peterson(N);     

                               // the same Peterson's algorithm (i.e., object)
                               // is passed to the N processes
      for (int i=0; i<N; i++) {
                               // i (= 0, ..., N-1) is the id of the process
         t[i] = new PetersonNProcesses(i,peterson); 
         t[i].start();                            
      }
   }
}
/** input:                    output:
 * ---------------------------------------------------------------------------
 *  javac PetersonNProcesses.java 
 *  java  PetersonNProcesses 3
 *                             0 1 2 [0] 0 [0] 0 [0] 0 [1] 1 [0] 0 [0] 0 [2] 2
 *                             [0] 0 [1] 1 [0] 0 [0] 0 [0] 0 [1] 1 [0] 0 [2] 2
 *                             [0] 0 ...
 *
 * ---------------------------------------------------------------------------
 *
 * Notice that the number of [0]'s is higher than the number of [1]'s which, 
 * in turn, is higher than the number of [2]'s. This is due to the fact that
 * the duration of the of the non-Critical Section is longer for processes 
 * with higher id.
 * ---------------------------------------------------------------------------
 */
