/**
 * ==========================================================================
 *                  Main Program for Testing Mutual Exclusion
 *
 * Thanks to Prof. Vijay K. Garg, Austin, Texas.
 * This program works for any implementation of the Lock interface
 * and it may test the mutual exclusion property among N processes: 
 * P_0, P_1, ..., P_{N-1}.
 * 
 * The Lock is implemented using Peterson's algorithm for 2 processes:
 * P_0 and P_1. In order to test Peterson's algorithm for 2 processes, 
 * we type:
 * 
 *    javac MutexTester.java
 *    java  MutexTester 2
 
 * For i = 0 or 1, in the output:
 *   i  means that process P_i is in its non-critical section
 *  [i] means that process P_i is in its     critical section
 * ==========================================================================
 */

import java.util.Random;

public class MutexTester extends Thread {
   int id;
   Lock lock;
   Random r = new Random();

   public MutexTester (int id, Lock lock) {   
      this.id = id;                            // id of process P_0 is 0
      this.lock = lock;                        // id of process P_1 is 1
   } 

   private static void Sleep(int time){
      try {
           Thread.sleep(time);
      }
      catch (InterruptedException e) {}
   }

   void nonCriticalSection() {
      System.out.print(" "+ id);
      Sleep(r.nextInt(300) * id + 50); // process P_0 takes 50 ms
   }                                   // process P_1 takes 50+random(300) ms

   void CriticalSection(){
      System.out.print(" ["+ id);
      Sleep(r.nextInt(300) * id + 50); // process P_0 takes 50 ms
      System.out.print("]");           // process P_1 takes 50+random(300) ms
   }

   public void run() {
      while (true) {
         nonCriticalSection();
         lock.takeLock(id);
         CriticalSection();
         lock.releaseLock(id);
      }
   }

   public static void main(String [] args) {
                                               // N is the number of processes
      final int N = Integer.parseInt(args[0]); 
                                               // t is an array of N threads
      MutexTester [] t = new MutexTester[N];      
                              // only one lock, i.e., peterson, is constructed
      Lock petersonTwoProcesses = new PetersonTwoProcesses();         
                              // the same lock is passed to the two processes
      for (int i=0; i<N; i++) {
                              // i (= 0, 1) is the id of the process
         t[i] = new MutexTester(i,petersonTwoProcesses);    
         t[i].start();                         
      }
   }
}
/** input:                output:
 * ---------------------------------------------------------------------------
 *  javac MutexTester.java 
 *  java  MutexTester 2
 *                         0 1 [0] 0 [1] 1 [0] 0 [0] 0 [1] 1 [0] 0 [1] 1 ...
 * ---------------------------------------------------------------------------
 */
